//------------------------------------------------------------------------------//
//                                                                              //
//  WInfo Copyright (C) Glenn Rosenthal, 2000-2001,2002.  All rights reserved.  //
//                                                                              //   
//  This utility program processes information from several soruces             //
//  to generate kana->kanji conversion dictionaries used by JWPce.              //
//                                                                              //
//  The main input data files come from the WNN package which was               //
//  deleveloped by Kyoto Universinity, OMRON, and ASTEC.  Please                //
//  See the copyright notice below.                                             //
//                                                                              //
//------------------------------------------------------------------------------//

//-------------------------------------------------------------------//
//
//  Copyright notice for WNN files (most of the .u files)       
//
//    Copyright Kyoto University Research Institute for Mathematical 
//            Sciences 1987, 1988, 1989, 1990, 1991
//    Copyright OMRON Corporation. 1987, 1988, 1989, 1990, 1991
//    Copyright ASTEC, Inc. 1987, 1988, 1989, 1990, 1991
//
//    Permission to use, copy, modify, distribute, and sell this software and 
//    its documentation for any purpose is hereby granted without fee, provided 
//    that all of the following conditions are satisfied:
//
//        1. The above copyright notices appear in all copies
//        2. Both those copyright notices and this permission notice appear 
//            in supporting documentation
//        3. The name of "Wnn" isn't changed unless substantial modifications 
//            are made, or the following words followed by the above copyright 
//            notices appear in all supporting documentation of software based 
//            on "Wnn":
//
//            "This software is based on the original version of Wnn developed 
//            by Kyoto University Research Institute for Mathematical Sciences 
//            (KURIMS), OMRON Corporation and ASTEC Inc."
//
//        4. The names KURIMS, OMRON and ASTEC may not be used in advertising 
//            or publicity pertaining to distribution of the software without 
//            specific, written prior permission of KURIMS, OMRON and ASTEC, 
//            who make no representations about the suitability of this software 
//            for any purpose. It is provided "as is" without express or implied 
//            warranty.
//
//  Wnn consortium is one of distributors of the official Wnn source code 
//  release. Wnn consortium also makes no representations about the suitability 
//  of this software for any purpose. It is provided "as is" without express or 
//  implied warranty.
//
//  KURIMS, OMRON, ASTEC and Wnn consortium disclaim all warranties with regard 
//  to this software, including all implied warranties of merchantability and 
//  fitness. In no event shall KURIMS, OMRON, ASTEC or Wnn consortium be liable 
//  for any special, indirect or consequential damages or any damages whatsoever 
//  resulting from loss of use, data or profits, whether in an action of contract, 
//  negligence or other tortuous action, arising out of or in connection with the 
//  use or performance of this software.
//
//-------------------------------------------------------------------//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <memory.h>
#include "winfo.h"

//-------------------------------------------------------------------//
//                                                                   //
//  Compile time options.                                            //
//                                                                   //
//-------------------------------------------------------------------//

//efine CHECK_DIC           // Causes a check to be done on dictionary files to make sure we know what the 
                            //   keys are.  This compairs key words that begin with a or v against a list
                            //   to pick of adj or verb codes that need processing.  This check causes an
                            //   error if there are any flags that are not known.  Generally you can disable
                            //   this.  I run each new version of EDICT through once with this to check for 
                            //   any new keywords.

#define DEBUG               // Enables the test file generation.  This generates 4 output files one at 
                            //   each stage of the process:
                            //      test1.euc -- Raw list data, unsorted and unfiltered.
                            //      test2.euc -- Sorted list.  Still unfiletered so should have duplciates.
                            //      test3.euc -- Filtered list.  Duplicate entries have been removed.
                            //      test4.euc -- Filan result.  Filtered, sorted, and merged.
                            

//--------------------------------
//
//  Test to see if a string matches the dictionary key.  To match the strings must be the same, and
//  the the string must be followed by ) or ,.
//
//      buffer -- Pointer to dictionary string.
//      string -- String to be conpiared.
//
//      RETURN -- Non-zero indicates a match.
//
static int test_dic (byte *buffer,char *string) {
  while (*string) {
    if (tolower(*string) != tolower(*buffer)) return (false);
    buffer++;
    string++;
  }
  if ((*buffer == ',') || (*buffer == ')')) return (true);
  return (false);
}

//--------------------------------
//
//  Tests a dictionary word to see if it is in the list of known keys or excluded words.
//
//      buffer -- Pointer to word to be tested.
//
//      RETURN -- Non-zero indicates the word is in the exclusion list.
//
static int test_dic (byte *buffer) {
  #define SIZE_LIST   (sizeof(list)/sizeof(char *))
  static char *list[] = {
    "abbr"  ,"adv"   ,"adj-na","adj-no","adj-pn","ateji" ,"vulg"  ,"arch"  ,"a-no"  ,"away"  ,"verse" ,"at"    ,"ation" ,"ASCII" ,"audio" ,"asr"   ,"ai:"   ,"auth"  ,"area"  ,"army"  ,"all"   ,"air"   ,"aid"   ,"also"  ,
    "about" ,"al"    ,"Ainu"  ,"ATM"   ,"anger" ,"armed" ,"among" ,"ACK"   ,"an"    ,"are"   ,"armor" ,"adopt" ,"asset" ,"arch." ,"ADF"   ,"ATC"   ,"ATS"   ,"very"  ,"abbr." ,"vital" ,"virus" ,"act"   ,"actor" ,"ahead" ,"AI"    ,"alive" ,
    "A.D."  ,"view"  ,"azure" ,"avid"  ,"Aesop" ,"and"   ,"voice" ,"ASEAN" ,"arbor" ,"anat." ,"votes" ,"ally"  ,"vice"  ,"award" ,"above" ,"anim"  ,"amend" ,"agent" ,"as"    ,"a"     ,"areas" ,"age"   ,"vine"  ,"valid" ,"atic"  ,"along" ,
    "vocative" ,"athletics","animal"   ,"anemia"   ,"anxiously","alcohol"  ,"anomaly"  ,"affairs"  ,"aburage"  ,"arrows"   ,"archery"  ,"artistic" ,"vegetable","alphabets","ancient"  ,"aircraft" ,"anti-war" ,"archaic"  ,"verbatim" ,
    "visual"   ,"associate","animate"  ,"American" ,"archeol"  ,"vendor"   ,"attitude" ,"villages" ,"vitamin-" ,"argument" ,"atomic"   ,"vessel"   ,"verbal"   ,"assured"  ,"animals"  ,"action"   ,"atmos."   ,"violin"   ,"variety"  ,
    "village"  ,"accounts" ,"assets"   ,"ability"  ,"author"   ,"actual"   ,"actress"  ,"attending","assembly" ,"assistant","arith."   ,"victory"  ,"article"  ,"all-night","veracity" ,"Ailanthus","annual"   ,"analysis" ,"agency"   ,
    "answer"   ,"atozoon"  ,"against"  ,"awareness","volume"   ,"aquarium" ,"aspects"  ,"ashore"   ,"another"  ,"anatomy"  ,"vermiform","adding"   ,"Akihabara","across"   ,"vulgar"   ,"around"   ,"Aikikai"  ,"acting"   ,"accepting",
    "arctic"   ,"attempt"  ,"awkward"  ,"attached" ,"attentive","adverb"   ,"vengeful" ,"violently","account"  ,"airplane" ,"Vietnam"  ,"adopted"  ,"attain"   ,"academic" ,"accident" ,"adjacent" ,"antenna"  ,"aesthetic","vassal"   ,
    "accomplished"  ,"aeronautical"  ,"vegetables"    ,"adjournment"   ,"appearance"    ,"vote-getting"  ,"advantaged"    ,"accounting"    ,"astringent"    ,"air-traffic"   ,"applicants"    ,"association"   ,"authoritative" ,"authorized"    ,
    "vegetation"    ,"apprentice"    ,"astronomical"  ,"automobile"    ,"appointment"   ,"authorities"   ,"assembling"    ,"apportioning"  ,"volleyball"    ,"atmospheric"   ,"announcement"  ,"automotive"    ,"additional"    ,"agriculture"   ,
    "administration","ammunition"    ,"auspicious"    ,
  };
  int i;
  for (i = 0; buffer[i] && (buffer[i] != ',') && (buffer[i] != ')'); i++) {
    if (buffer[i] == ' ' ) return (true);
    if (buffer[i] == '\'') return (true);
  }
  if ((buffer[0] == 'a') && (buffer[1] == ' ')) return (true);
  for (i = 0; i < SIZE_LIST; i++) {
    if (test_dic(buffer,list[i])) return (true);
  }
  return (false);
}

//-------------------------------------------------------------------//
//                                                                   //
//  KANJI_index class.                                               //
//                                                                   //
//  This class defines an entry in the WNN index file.               //
//                                                                   //
//-------------------------------------------------------------------//

//--------------------------------
//
//  Clear the index control
//
void KANJI_index::clear () {
  key[0] = key[1] = key[2] = key[3];
  offset = 0;
  return;
}

//-------------------------------------------------------------------//
//                                                                   //
//  WNN_node class.                                                  //
//                                                                   //
//  This class defines a node.  Initally, this is a single entry     //
//  into the conversion list.  After the merge, this may represent   //
//  several possible entries.                                        //
//                                                                   //
//-------------------------------------------------------------------//

//--------------------------------
//
//  Constructor.
//  
WNN_node::WNN_node () {
  kana[0]   = kanji[0] = 0;     // No data
  key[0] = key[1] = key[2] = 0; // No key
  godan     = GODAN_NONE;       // Not a godan verb
  wnn_pri   = 0;                // Zero WNN priority (lowest)
  kanji_pri = 5000;             // 5000 kanji priority (highest)
  type      = TYPE_NOUN;        // Not a godan verb.
  len       = 0;                // No kanji.
  return;
}

//--------------------------------
//
//  This constructor generates a node based on another node.
//
//      node -- Node to base the new node on.
//      end  -- If non-zero this is a character to add to the end of the conversion.
//
//  In reallity, this was added to handle irregular verbs such as gozaru which becomes
//  gozai(masu).  Adding this conversion will set the conversion status of the new 
//  conversion to a noun type.
//  
WNN_node::WNN_node (WNN_node *node,KANJI end) {
  int i;
  *this = *node;
  if (end) {
    for (i = 0; kana[i]; i++);
    kana [i    ] = end;
    key  [i++  ] = end & 0x7f;
    kana [i    ] = 0;
    key  [i    ] = 0;
    kanji[len++] = end;
    kanji[len  ] = 0;
    godan        = 0;
    type         = TYPE_NOUN;
  }
  return;
}

//--------------------------------
//
//  Constructor from a node from a WNN data line.
//
//  The bassic format of a WNN line is (in EUC):
//
//      kana<tab>kanji<tab><part of speach><tab><wnn_priority>
//
//  If the entry is not to be kept, the return value will be TYPE_DELETE.  Generally, 
//  this type is created by entries that don't contain any kanji
//
//  In the case of special verb endings gozaru->gozaimasu, this routine will return a 
//  TYPE_SPECIAL.  The reciever is expected to change this to  TYPE_GODAN, and generate
//  a TYPE_NOUN entry with the i ending.
//
#define ISOK(x)     (!ISKANA((x)) && ((x) != KANJI_LONGVOWEL))
#define ISEND(x)    ((x == '\t') || (x == ' ') || (x == KANJI_SPACE))

WNN_node::WNN_node (byte *buffer) {
  int   i,j;
  KANJI part[100];
  for (i = 0; !ISEND(*buffer); i++, buffer += 2) {              // Get the kana and geneate the key
    kana[i] = ((buffer[0] << 8) | buffer[1]) & 0x7f7f;
    key [i] = kana[i] & 0x7f;
  } 
  kana[i  ] = 0;
  key [i  ] = 0;
  key [i+1] = 0;                                                // Takes care of one character kana's
  while (ISEND(*buffer)) buffer++;                              // Skip tabs then get the kanji.
  for (i = 0; !ISEND(*buffer); i++, buffer += 2) kanji[i] = ((buffer[0] << 8) | buffer[1]) & 0x7f7f;
  kanji[i] = 0;
  len      = i;
  for (j = i = 0; kanji[i]; i++) if (ISOK(kanji[i])) j++;       // If the kanji string does not contain any
  if (!j) { type = TYPE_DELETE; return; }                       //   kanji, we will skip this entry.
  while (ISEND(*buffer)) buffer++;
  for (i = 0; !ISEND(*buffer); i++) {                           // Get part of speach.  Note this string
    if (buffer[0] & 0x80) {                                     //   can contain ascii so has to be decodeed
      part[i] = ((buffer[0] << 8) | buffer[1]) & 0x7f7f;        //   as full EUC.
      buffer += 2;
    }
    else {
      part[i] = buffer[0];
      buffer++;
    }
  }
  part[i] = 0;
  while (ISEND(*buffer)) buffer++;                              // Skip tabs
  sscanf ((char *) buffer,"%d",&wnn_pri);                       // Get WNN priority (higher is a more common word).
  godan   = GODAN_NONE;
  switch (part[0]) {                                            // Determine important parts of speach.
    case 0x306C:        // Ichi-dan verb
         type  = TYPE_ICHIDAN;
         godan = '1';
         break;
    case 0x3741:        // Adjitive
         type  = TYPE_IADJ;
         godan = 'i';
         break;
    case 0x256F:        // WA (U) verb;
         type  = TYPE_GODAN;
         godan = 'u';
         break;
    case 0x2569:        // RA (RU) verb;
         type  = TYPE_GODAN;
         godan = 'r';                               // Check for irregular verb eding gozaru->gozaimasu
         if ((part[3] == 0x323C) && (part[4] == 0x2435) && (part[5] == 0x2424)) type = TYPE_SPECIAL;
         break;
    case 0x252B:        // KA (KU) verb;
         type  = TYPE_GODAN;
         godan = 'k';
         break;
    case 0x255E:        // MA (MU) verb;
         type  = TYPE_GODAN;
         godan = 'm';
         break;
    case 0x252C:        // GA (GU) verb;
         type  = TYPE_GODAN;
         godan = 'g';
         break;
    case 0x253F:        // TA (TU) verb;
         type  = TYPE_GODAN;
         godan = 't';
         break;
    case 0x2550:        // BA (BU) verb;
         type  = TYPE_GODAN;
         godan = 'b';
         break;
    case 0x2536:        // ZA (ZU) verb;
         type  = TYPE_GODAN;
         godan = 'z';
         break;
    case 0x254A:        // NA (NU) verb;
         type  = TYPE_GODAN;
         godan = 'n';
         break;
    case 0x2535:        // SA (SU) verb;
         if ((part[3] == 0x2439) && (part[4] == 0x246B)) {  // This is actually a suru (VS) noun
           type = TYPE_NOUN;
           break;
         }
         type  = TYPE_GODAN;
         godan = 's';
         break;
    default:
         printf ("\n\nWARNING: unknown kanji type [%x]\n",part[0]);
    case 0x3847:        // Personal pro-noun;
    case 0x405C:        // Prefix
    case 0x3D75:        // Postfix, counter
    case 0x3F74:        // Number
    case 0x4C3E:        // Noun
    case 0x497B:        // Adverb
    case 0x4f22:        // pre-noun adjectival
    case 0x4331:        // Single kanji (kanji reading)
    case 0x434f:        // Place name
    case 0x3f4d:        // Personal name
    case 0x4643:        // Special entries I generated
    case 0x4d68:        // Some wierd from of kuru
    case 0x352D:        // Symbol
         type = TYPE_NOUN;
         break;
  }  
  return;
}

//--------------------------------
//
//  Constructor from a node from a WNN data line.
//
//  The bassic format of a DIC line is (in EUC):
//
//      kanji [<kana>] /meaning/
//
//  If the entry is not to be kept, the return value will be TYPE_DELETE.  Generally, 
//  this type is created by entries that don't contain any kanji
//
//  In the case of special verb endings gozaru->gozaimasu, this routine will return a 
//  TYPE_SPECIAL.  The reciever is expected to change this to  TYPE_GODAN, and generate
//  a TYPE_NOUN entry with the i ending.
//
//      buffer   -- Buffer containning dictionary line to be anlayzed.
//      priority -- If non-zero only priority words will be accepted.
//
#define SIZE_CODE   20

WNN_node::WNN_node (byte *buffer,int priority) {
  int   i,j,k,intype,verb,vs;
  byte *meaning;
//
//  Get the first string.  This should be kanji.  If it is just kana we don't need this one.
//
  for (i = 0; *buffer != ' '; buffer += 2, i++) {
    kanji[i] = ((buffer[0] << 8) | buffer[1]) & 0x7f7f;
  }
  kanji[i] = 0;
  len      = i;
  buffer++;
  if (*buffer != '[') { type = TYPE_DELETE; return; }
  buffer++;
//
//  Get the kana string.  If this has a long vowel, we can drop it.
//
  for (i = 0; *buffer != ']'; buffer += 2, i++) {
    kana[i] = ((buffer[0] << 8) | buffer[1]) & 0x7f7f;
    key [i] = kana[i] & 0x7f;
    if (kana[i] == KANJI_LONGVOWEL) { type = TYPE_DELETE; return; }
  }
  kana[i] = 0;
  key [i] = 0;
  wnn_pri = 0;
  meaning = buffer;
  meaning[strlen((char *) meaning)-1] = 0;
//
//  Check part of speach.
//
  godan   = GODAN_NONE;
  type    = TYPE_NOUN;
  verb    = false;
  vs      = false;
  for (intype = false; *buffer; buffer++) {
    switch (*buffer) {
      case '(':
           intype = true;
           break;
      case ')':
           intype = false;
           continue;
      case ',':
           if (!intype) continue;
           break;
       default:
           continue;
    }
    buffer++;
//
//  Pick off priority entries.
//
    if      (test_dic(buffer,"P"     )) wnn_pri = 1;
//
//  These need more work.
//
    else if (test_dic(buffer,"vk"    )) { type = TYPE_ICHIDAN; }                        // kuru type special verb.
    else if (test_dic(buffer,"v5k-s" )) {                                               // iku/yuku type special verb.
      for (i = 0; kanji[i]; i++);
      for (j = 0; kana [j]; j++);
      for (k = 0; key  [k]; k++);
      i--; j--; k--; len--;
      for (i = 0; kanji[i]; i++);
      for (j = 0; kana [j]; j++);
      for (k = 0; key  [k]; k++);
      i--; j--; k--; len--;
    }
//
//  These may be improved at a later time.
//
    else if (test_dic(buffer,"v1"    )) dic_verb (KANJI_RU, 0 ,TYPE_ICHIDAN,meaning);   // ichi-dan verb
    else if (test_dic(buffer,"v5uru" )) dic_verb (KANJI_RU,'r',TYPE_GODAN  ,meaning);   // Godan verb - Uru old class verb (old form of Eru)
    else if (test_dic(buffer,"v5z"   )) dic_two  (KANJI_ZU,KANJI_RU,meaning);           // Godan verb - -zuru special class (alternative form of -jiru verbs)
    else if (test_dic(buffer,"vs-s"  )) dic_two  (KANJI_SU,KANJI_RU,meaning);           // suru verb - special class
    else if (test_dic(buffer,"adj-t" )) dic_two  (KANJI_TA,KANJI_RU,meaning);           // adj-t	   "taru" adjective
    else if (test_dic(buffer,"v5"    )) {                                               // Unspecified verb, will assume is a godan
      for (i = 0; kanji[i]; i++);
      i--;
      switch (kanji[i]) {
        case KANJI_U : dic_verb (KANJI_U ,'u',TYPE_GODAN,meaning); break;
        case KANJI_KU: dic_verb (KANJI_KU,'k',TYPE_GODAN,meaning); break;
        case KANJI_GU: dic_verb (KANJI_GU,'h',TYPE_GODAN,meaning); break;
        case KANJI_SU: dic_verb (KANJI_SU,'s',TYPE_GODAN,meaning); break;
        case KANJI_TU: dic_verb (KANJI_TU,'t',TYPE_GODAN,meaning); break;
        case KANJI_DU: dic_verb (KANJI_DU,'d',TYPE_GODAN,meaning); break;
        case KANJI_NU: dic_verb (KANJI_NU,'n',TYPE_GODAN,meaning); break;
        case KANJI_FU: dic_verb (KANJI_FU,'f',TYPE_GODAN,meaning); break;
        case KANJI_BU: dic_verb (KANJI_BU,'b',TYPE_GODAN,meaning); break;
        case KANJI_PU: dic_verb (KANJI_PU,'p',TYPE_GODAN,meaning); break;
        case KANJI_MU: dic_verb (KANJI_MU,'m',TYPE_GODAN,meaning); break;
        case KANJI_RU: dic_verb (KANJI_RU,'r',TYPE_GODAN,meaning); break;
        default:
             printf ("ERROR -- Invalid ending for verb [%s]\n",meaning);
             type = TYPE_ERROR;
             break;
      }
    }
//
//  These are fine.
//
    else if (test_dic(buffer,"v5aru" )) dic_verb (KANJI_RU,'r',TYPE_SPECIAL,meaning);   // Godan verb - -aru special class (gozaru -> gozaimasu)
    else if (test_dic(buffer,"adj"   )) dic_verb (KANJI_I ,'i',TYPE_IADJ   ,meaning);   // i-adjective
    else if (test_dic(buffer,"v5u"   )) dic_verb (KANJI_U ,'u',TYPE_GODAN  ,meaning);   // Godan verbs
    else if (test_dic(buffer,"v5k"   )) dic_verb (KANJI_KU,'k',TYPE_GODAN  ,meaning);   
    else if (test_dic(buffer,"v5g"   )) dic_verb (KANJI_GU,'g',TYPE_GODAN  ,meaning);
    else if (test_dic(buffer,"v5s"   )) dic_verb (KANJI_SU,'s',TYPE_GODAN  ,meaning);
    else if (test_dic(buffer,"v5t"   )) dic_verb (KANJI_TU,'t',TYPE_GODAN  ,meaning);
//  else if (test_dic(buffer,"v5d"   )) dic_verb (KANJI_DU,'d',TYPE_GODAN  ,meaning);
    else if (test_dic(buffer,"v5n"   )) dic_verb (KANJI_NU,'n',TYPE_GODAN  ,meaning);
//  else if (test_dic(buffer,"v5f"   )) dic_verb (KANJI_FU,'f',TYPE_GODAN  ,meaning);
    else if (test_dic(buffer,"v5b"   )) dic_verb (KANJI_BU,'b',TYPE_GODAN  ,meaning);
//  else if (test_dic(buffer,"v5p"   )) dic_verb (KANJI_PU,'p',TYPE_GODAN  ,meaning);
    else if (test_dic(buffer,"v5m"   )) dic_verb (KANJI_MU,'m',TYPE_GODAN  ,meaning);
    else if (test_dic(buffer,"v5r"   )) dic_verb (KANJI_RU,'r',TYPE_GODAN  ,meaning);
//
//  These generate no input.
//
    else if (test_dic(buffer,"aux-v" )) {  }
    else if (test_dic(buffer,"vs"    )) vs   = true; 
    else if (test_dic(buffer,"vi"    )) verb = true;
    else if (test_dic(buffer,"vt"    )) verb = true;
//
//  This is the pickup for anything left.  
//
//  If trhe dictionary check is on, this will require that we know these keywords.
//
    else {
#ifdef CHECK_DIC
      if (((tolower(*buffer) == 'a') || (tolower(*buffer) == 'v')) && !test_dic(buffer)) {
        char *ptr,buf[512];
        strcpy (buf,(char *) buffer);
        for (ptr = buf; *ptr && (*ptr != ',') && (*ptr != ')'); ptr++);
        *ptr = 0;
        if (strlen(buf) < SIZE_CODE) {
          printf ("ERROR -- Unknown dictionary code [%s]\n",buf);
          type = TYPE_ERROR;
        }
      }
#endif CHECK_DIC
    }
  }
  if (priority && (wnn_pri != 1)) type = TYPE_DELETE;                       // Do we want only priority words.
  if (verb && !vs && (type == TYPE_NOUN)) {                                 // Check that vi and vt are watched with a real verb.
    printf ("WARNING -- Verb indicated by not declared [%s]\n",meaning);
  }
  return;
}

//--------------------------------
//
//  Apends a character to the kanji buffer
//
//      k      -- Character to append.
//
//      RETURN -- A non-zero value indicates an error.  Basically buffer over-run is the 
//                only error possible.
//
int WNN_node::add (KANJI k) {
  if (len >= SIZE_KANJI-1) return (true);
  kanji[len++] = k;
  kanji[len  ] = 0;
  return (false);
}

//--------------------------------
//
//  This is a debugging routine used to write the node information to a file.
//
//      file -- File to write to.
//
void WNN_node::debug (FILE *file) {
#ifdef DEBUG
  int i;
  for (i = 0; kana [i]; i++) fprintf (file,"%c%c",(kana [i]>>8)|0x80,(kana [i]&0xff)|0x80);
  fprintf (file,"\t%c\t",godan ? godan : ' ');
  for (i = 0; kanji[i]; i++) {
    if (kanji[i] < 0x80) fprintf (file,"%c",(char) kanji[i]); else fprintf (file,"%c%c",(kanji[i]>>8)|0x80,(kanji[i]&0xff)|0x80);
  }
  fprintf (file,"\t%d\n",wnn_pri);
#endif DEBUG  
  return;
}

//--------------------------------
//
//  This routine modifies a node to remove the last two kana, and change the 
//  type of the node to ichidan.  This is used for taru and suru verbs which 
//  already have basic conjugation.  Generally this should be descuraged.
//
//      end1    -- First ending kana.
//      end2    -- Second ending kana.
//      meaning -- Used for error generation.
//
//      RETURN  -- Non-zero value idnicates and error.
//
//  Expected pattern is:
//
//      <kanji/kana><end1><end2>
//
int WNN_node::dic_two (KANJI end1, KANJI end2,byte *meaning) {
  int i,j,k;
  for (i = 0; kanji[i]; i++);                                                       // Find end of all strings.
  for (j = 0; kana [j]; j++);
  for (k = 0; key  [k]; k++);
  i--; j--; k--; len--;                                                             // Backup one character.
  if ((end2 != kanji[i]) || (end2 != kana[j]) || ((end2 & 0x00ff) != key[k])) {     // Check for end2.
    printf ("ERROR -- Ending does not match the actual verb [- %s]\n",meaning);
    type = TYPE_ERROR;
    return (true);
  }
  i--; j--; k--; len--;                                                             // Backup another character.
  if ((end1 != kanji[i]) || (end1 != kana[j]) || ((end1 & 0x00ff) != key[k])) {     // Check for end1
    printf ("ERROR -- Ending does not match the actual verb [- %s]\n",meaning);
    type = TYPE_ERROR;
    return (true);
  }
  kanji[i] = 0;                                                                     // Terminate strings, and set valeus.
  kana [j] = 0;
  key  [k] = 0;
  godan    = GODAN_NONE;
  type     = TYPE_ICHIDAN;
  return (false);
}

//--------------------------------
//
//  Configures a verb ending for a node.
//
//  The last kana must match the end, which is removed and the type and godan are changed 
//  as indicated.
//
//      end     -- End kana to be removed.
//      _godan  -- Godan type to be set (can be GODAN_NONE).
//      _type   -- Verb type (TYPE_ICHIDAN, TYPE_GODAN, TYPE_IADJ)
//      meaning -- Used for error messages.
//
//      RETURN  -- Non-zero value indicates an error.
//
int WNN_node::dic_verb (KANJI end,int _godan,int _type,byte *meaning) {
  int i,j,k;
  for (i = 0; kanji[i]; i++);
  for (j = 0; kana [j]; j++);
  for (k = 0; key  [k]; k++);
  i--; j--; k--; len--;
  if ((end != kanji[i]) || (end != kana[j]) || ((end & 0x00ff) != key[k])) {
    printf ("ERROR -- Ending does not match the actual verb [%c %s]\n",_godan ? _godan : '=',meaning);
    type = TYPE_ERROR;
    return (true);
  }
  kanji[i] = 0;
  kana [j] = 0;
  key  [k] = 0;
  godan    = _godan;
  type     = _type;
  return (false);
}

//--------------------------------
//
//  Determines if two nodes are identical!
//
//      node   -- Node to compare with this one.
//
//      RETURN -- Indicates the state of the match.
//
//                  MATCH_MATCH -- Indicates full match.
//                  MATCH_PART  -- Indicates the kana and godan match.
//                  MATCH_NONE  -- Indicates the nodes do not match
//
int WNN_node::match (WNN_node *node) {
  int i;
  for (i = 0; key[i] || node->key[i]; i++) {            // Test the keys.
    if (key[i] != node->key[i]) return (MATCH_NONE);
  }
  if (godan != node->godan) return (MATCH_NONE);        // Test the godan type.
  for (i = 0; kanji[i] || node->kanji[i]; i++) {        // Test the kanji.
    if (kanji[i] != node->kanji[i]) return (MATCH_PART);
  }
  return (MATCH_ALL);
}

//--------------------------------
//
//  Appends a node to this one.  This bassicaly copies the kanji out of the node and sticks
//  it on the end of this one.
//
//      node   -- Node to copy from.
//
//      RETURN -- A non-zero value indicates an error (basically a buffer over-run).
//
int WNN_node::merge (WNN_node *node) {
  KANJI *k;
  if (add('/')) return (true);
  for (k = node->kanji; *k; k++) {
    if (add(*k)) return (true);
  }
  return (false);
}

//--------------------------------
//
//  Determines if two nodes are mergable (eg have same key and godan)!
//
//      node   -- Node to compare with this one.
//
//      RETURN -- A non-zero return indicaes the nodes are mergable.
//
int WNN_node::mtest (WNN_node *node) {
  int i;
  for (i = 0; key[i] || node->key[i]; i++) {            // Test the keys.
    if (key[i] != node->key[i]) return (false);
  }
  if (godan != node->godan) return (false);             // Test the godan type.
  return (true);
}

//--------------------------------
//
//  This routine compairs two nodes to determine what order they go in.  
//
//  Comparison is first based on the key value.  If the keys are equal, the godan type 
//  is then compaired.  Finally, the priority is compared.  The end result is that 
//  conversions are in order by type then the highest priority is placed first.
//
//      node   -- Node to compair with this one.
//
//      RETURN -- Indcates the node compairisons.
//
//          < 0 -- node comes before this one.
//          > 0 -- node comes after this one.
//          = 0 -- node are equivalent.  Note this does not mean the nodes 
//                 are the same, just they come in the same spot in the order.
//
int WNN_node::test (WNN_node *node) {
  int i;
  for (i = 0; (key[i] || node->key[i]) && (key[i] == node->key[i]); i++);
  if (node->key[i] == key[i]) i = node->godan-godan; else i = node->key[i]-key[i];
  if (i) return (i);
  return (wnn_pri-node->wnn_pri);
}

//--------------------------------
//
//  This routine determines if the current index key is different than the key for
//  this node.  Note, index keys contain only 3 kana.
//
//      index  -- Pointer to KANJI_index structure that describes the current
//                index value.
//
//      RETURN -- A non-zero return indicates the key has changed.  The key in 
//                index will be changed, and the index should be written to 
//                the file.
//
int WNN_node::test_key (KANJI_index *index) {
  int i;
  for (i = 0; (i < 3) && (key[i] == (index->key[i] & 0x7f)); i++);
  if (i == 3) return (false);
  for (i = 0; i < 3; i++) index->key[i] = (key[i] | 0x80);
  return (true);
}

//--------------------------------
//
//  This rutine writes a node to the dictionary file.
//
//      file   -- File to write to.
//
//      RETURN -- Number of bytes written.  A value of zero is returned for errors.
//
#define PUT(c)  { fputc ((c),file); len++; }    // Put character and increment length.

int WNN_node::write (FILE *file) {
  int i;
  int len = 0;                                  // Nothing written yet.
  for (i = 0; key[i]; i++) PUT(key[i] | 0x80);  // Write key.
  PUT('\t');                                    // Write tab
  PUT(godan ? godan : '*');                     // Write godan indicator
  PUT('\t');                                    // Write tab
  for (i = 0; kanji[i]; i++) {                  // Write kanji
    if (kanji[i] < 0x80) { PUT(kanji[i]) }
      else {
        PUT((kanji[i]>> 8   ) | 0x80);
        PUT((kanji[i] & 0xff) | 0x80);
      }
  }
  PUT    ('\r');                                // Write end of line.
  PUT    ('\n');
  return (len); 
}

//--------------------------------
//
//  This rutine writes a node to the dictionary file in new format..
//
//      file   -- File to write to.
//
//      RETURN -- Number of bytes written.  A value of zero is returned for errors.
//
int WNN_node::write_new (FILE *file) {
  int i;
  int len = 0;                                  // Nothing written yet.
  for (i = 0; key[i]; i++) PUT(key[i] | 0x80);  // Write key.
  PUT(godan ? godan : '*');                     // Write godan indicator
  for (i = 0; kanji[i]; i++) {                  // Write kanji
    if (kanji[i] < 0x80) { PUT(kanji[i]) }
      else {
        PUT((kanji[i]>> 8   ) | 0x80);
        PUT((kanji[i] & 0xff) | 0x80);
      }
  }
  PUT    ('\n');                                // Write end of line.
  return (len); 
}

//-------------------------------------------------------------------//
//                                                                   //
//  WNN_index class.                                                 //
//                                                                   //
//  This class defines an entire conversion list.                    //
//                                                                   //
//-------------------------------------------------------------------//

//#define BINARY_SEARCH       // Determines if a crude binary search is used in node lists.
                            // This speeds up the program a little.

//--------------------------------
//
//  Constructor.
//
WNN_index::WNN_index (void) {
  list  = NULL;                 // No nodes
  count = 0;                    // No list.
  alloc = 0;
  return;
}

//--------------------------------
//
//  Deconstructor
//
WNN_index::~WNN_index (void) {
  int i;
  if (list) {
    for (i = 0; i < alloc; i++) {
      if (list[i]) delete (list[i]);
    }
    free (list);
  }
  list  = NULL;
  alloc = 0;
  count = 0;
  return;
}

//--------------------------------
//
//  This is a debuggin routine that simply prints out all of the nodes in the list to a file.
//
//      name   -- Name of file to wirte to.
//
//      RETURN -- A non-zero return indicates an error.
//
int WNN_index::debug (char *name) {
  int   i;
  FILE *file;
  if (!(file = fopen(name,"wb"))) {
    printf ("\n\nERROR: Cannot open file for writing %s\n",name);
    return (true);
  }
  for (i = 0; i < count; i++) list[i]->debug (file);
  fclose (file);
  return (false);
}

//--------------------------------
//
//  This routine removes duplicate entries from the list.  The entry with lower priority is 
//  automatically removed.
//
//  Note the list is sorted by kana then godan then priority so we cannot count on duplicates
//  being anywhere near each other in the list.
//
void WNN_index::filter (void) {
  int i,j,k;
  int dup = 0;
  printf ("Filtering duplicate entries ... ");
  for (i = 0; i < count; i++) {                                 // Move through list.
    for (j = i+1; j < count; ) {                                // Scan all entries with same kana and godan 
      switch (list[i]->match(list[j])) {                        // Test entry.
        case MATCH_ALL:                                         // Match so remove the second entry which will always be lower or equal priority.
             for (k = j; k < count; k++) list[k] = list[k+1];
             count--;
             dup++;
             break;
        case MATCH_PART:                                        // Godan and kana match, but not kanji so keep scanning.
             j++;                                               //   We can not count on the kanji being near each other because of the priority sort.
             break;
        case MATCH_NONE:                                        // Kana or godan did not match so lets get out of here.
             goto Next;
      }
    }
Next:;    
  }
  printf ("total=%d dup=%d\n",count,dup);
  return;
}

//--------------------------------
//
//  Finds a specific node if it exits.
//
//      node   -- Node to find.  Not this node, but an identical node.
//
//      RETURN -- Located node or a value of zero.
//
WNN_node *WNN_index::find (WNN_node *node) {
  int i;
  for (i = 0; i < count; i++) {
    if (list[i]->match(node)) return (list[i]);
  }
  return (NULL);
}

//--------------------------------
//
//  Intializes the index structure.  This is basically to allocate the list arrya.
//
//      size   -- Number of indexes to allocate (4 bytes each).
//
//      RETURN -- Non-zero indicates an error.
//
int WNN_index::init (long size) {
  if (!(list = (WNN_node **) calloc(size+5,sizeof(WNN_node *)))) return (true);
  alloc = size;
  return (false);
}

//--------------------------------
//
//  Inserts a node into the appropriate location in the list.  The insersion is based
//  first on sort by conversion by key, then godan parameter.  The final sorting parameter
//  is determined by priority of the kanji.  Higher priority kanji are placed in the list 
//  earlier.
//
//      node   -- Node to be inserted.
//
//      RETURN -- A non-zero return value indicates that this entry is actually a duplicate.
//
int WNN_index::insert (WNN_node *node) {
  list[count++] = node;
  if (count >= alloc) { printf ("\n\nERROR -- Too many entries allocate more list space!\n\n"); exit (1); }
  return (false);
}

//--------------------------------
//
//  This is a special routine for handling the case when a new conversion returned
//  TYPE_SPECIAL.  This is always a responce to an irregular verb that converts simular
//  to gozaru (which convertes to gozai[masu]). 
//
//  This routien adds a second entry, so there would be two entries:
//
//      goza    -> <kanji>      r-godan
//      gozai   -> <kanji>i     noun
//
//  This routine will change the type on the main node back to a godan.
//
//      node   -- Node for special item.
//
//      RETURN -- A non-zero return indicates an error.
//
#define KANA_I  0x2424

int WNN_index::irregular (WNN_node *node) {
  WNN_node *n;
  if (!(n = new WNN_node(node,KANA_I))) { printf ("\n\nERROR: Out of memory\n"); return (true); }
  if (insert(n)) delete n;
  node->type = TYPE_GODAN;
  return (false);
}

//--------------------------------
//
//  Merge all mergable entries in the list.  Before this is called, the list should be put 
//  in order by key, godan, and priority.  This routine will merge adjacent entries that 
//  contain identical keys and godan status to make a more compact working list.
//
//      RETURN -- A non-zero return indicates an error
//
int WNN_index::merge (void) {
  int i,j;
  printf ("merging entires with same key ... ");
  for (i = 0; i < count-1; i++) {                       // Move thought list.
    while ((i < count-1) && list[i]->mtest(list[i+1])) {
      if (list[i]->merge(list[i+1])) { printf ("\n\nERROR: Buffer overrun in data merge [buffer size %d]\n",SIZE_KANJI); return (true); }
      delete list[i+1];
      for (j = i+1; j < count; j++) list[j] = list[j+1];
      count--;
    }
  }
  printf ("done (%d entries)\n",count);
  return (false);
}

//--------------------------------
//
//  Quicksort used for sotring the list of keys.
//
//      low,high -- Bounds for the sort.  Call with arguments (0,count).
//
//  This is a recursive routine and can use lots of stack space.
//
#define SWAP(x,y) { temp = list[x]; list[x] = list[y]; list[y] = temp; }

void WNN_index::quicksort (long low,long high) {
  long      i,last;
  WNN_node *temp;
  if (low >= high) return;                          // Done.
  i    = (low+high)/2;                              // This is not necessary, but is done in case the list is in order (Quicksort does not
  SWAP (low,i);                                     //   work well when the list is in order, but this should not happen here.
  last = low;
  for (i = low+1; i <= high; i++) {                 // Partition.  Move all elements before selected (low, but was mid) element before the 
    if (list[low]->test(list[i]) < 0) {             //   partition marker.
      last++;
      SWAP (i,last);
    }
  }
  SWAP      (low,last);                             // Swap partion marker to end, between lists, since this element is were it belongs.
  quicksort (low,last-1);                           // Sort both halves of the list.
  quicksort (last+1,high);
  return;
}

//--------------------------------
//
//  Adds a single line entry to the file.  The entry must be in the same format as the 
//  WNN database.
//
//      buffer -- Line to add.
//  
//      RETURN -- A non-zero value idnicates an error.
//
//  The WNN data files are EUC files with the following format:
//
//      <kana><tab><kanji><tab><part of speech><tab><priority>  
//
int WNN_index::read_line (char *buffer) {
  WNN_node *node;
  printf ("Add Line ... ");
  if (!(node = new WNN_node((byte *) buffer))) { printf ("\n\nERROR: Out of memory\n"); return (true); }
  if (node->type == TYPE_DELETE) { delete node; return (false); }
  if ((node->type == TYPE_SPECIAL) && irregular(node)) return (true);
  if (insert(node)) { delete node; return (false); }
  printf ("total=%d\n",count);
  return (false);
}

//--------------------------------
//
//  This reads the entries out of a WNN data file.
//
//      name   -- Name of file to read.
//
//      REUTRN -- Non-zero value indicates an error.
//
//  The WNN data files are EUC files with the following format:
//
//      <kana><tab><kanji><tab><part of speech><tab><priority>  
//
int WNN_index::read_wnn (char *name) {
  int       lines   = 0;
  int       deleted = 0;
  int       dup     = 0;
  int       i       = count;
  WNN_node *node;
  FILE     *file;
  byte      buffer[512];

  printf ("WNN file: %s ... ",name);
  if (!(file = fopen(name,"r"))) { printf ("\n\nERROR: cannot open file %s\n",name); return (true); }
  while (true) {
    fgets ((char *) buffer,sizeof(buffer),file);
    if (feof(file)) break;
    lines++;
    if (!buffer[0] || (buffer[0] == '\\') || (buffer[0] == '\r') || (buffer[0] == '\n')) continue;
    if (!(node = new WNN_node(buffer))) { printf ("\n\nERROR: Out of memory\n"); return (true); }
    if (node->type == TYPE_DELETE) { deleted++; delete node; continue; }
    if ((node->type == TYPE_SPECIAL) && irregular(node)) return (true);
    if (insert(node)) { dup++; delete node; continue; }
  } 
  printf ("lines=%d ent=%d total=%d del=%d dup=%d\n",lines,count-i,count,deleted,dup);
  fclose (file);
  return (false);
}

//--------------------------------
//
//  This routine writes the dictionary and index file in old JWP format.
//
//      base   -- Base name to use for writing (dict=base+".dct" and index=base+."idx").
//
//      RETURN -- A non-zero return value indicates an error.
//
int WNN_index::write (char *base) {
  int          i;
  char         buffer[512];
  FILE        *idx,*dic;
  KANJI_index  index;
//
//  Setup and open the files.
//
  printf ("Writing dictionary and index ... ");
  strcpy (buffer,base);
  strcat (buffer,".dct");
  if (!(dic = fopen(buffer,"wb"))) { printf ("\n\nERROR: Unable to open dictionary file %s\n",buffer); return (true); }
  strcpy (buffer,base);
  strcat (buffer,".idx");
  if (!(idx = fopen(buffer,"wb"))) { printf ("\n\nERROR: Unable to open index file %s\n",buffer); fclose (dic); return (true); }
//
//  Output the files.
//
  index.clear ();
  for (i = 0; i < count; i++) {
    if (list[i]->test_key(&index)) fwrite (&index,sizeof(KANJI_index),1,idx);
    index.offset += list[i]->write (dic);
  }
//
//  Cleanup
//
  fclose (idx);
  fclose (dic);
  printf ("done!\n");
  return (false);
}

//--------------------------------
//
//  This routine writes the dictionary and index file in new JWPce format.
//
//      base   -- Base name to use for writing (dict=base+".dat" and index=base+."dix").
//
//      RETURN -- A non-zero return value indicates an error.
//
int WNN_index::write_new (char *base) {
  int          i;
  char         buffer[512];
  FILE        *idx,*dic;
  KANJI_index  index;
//
//  Setup and open the files.
//
  printf ("Writing dictionary and index in new format ... ");
  strcpy (buffer,base);
  strcat (buffer,".dat");
  if (!(dic = fopen(buffer,"wb"))) { printf ("\n\nERROR: Unable to open dictionary file %s\n",buffer); return (true); }
  strcpy (buffer,base);
  strcat (buffer,".dix");
  if (!(idx = fopen(buffer,"wb"))) { printf ("\n\nERROR: Unable to open index file %s\n",buffer); fclose (dic); return (true); }
//
//  Output the files.
//
  index.clear ();
  for (i = 0; i < count; i++) {
    if (list[i]->test_key(&index)) fwrite (&index,sizeof(KANJI_index),1,idx);
    index.offset += list[i]->write_new (dic);
  }
//
//  Cleanup
//
  fclose (idx);
  fclose (dic);
  printf ("done!\n");
  return (false);
}

//--------------------------------
//
//  Read a file in EDICT format and extract the kana->kanji conversions from it.
//
//      name     -- File to read.
//      priority -- If non-zero, only priority entries 
//
//      RETURN   -- Non-zero value indicates an error.
//
int WNN_index::read_dic (char *name,int priority) {
  int       lines   = 0;
  int       deleted = 0;
  int       dup     = 0;
  int       i       = count;
  WNN_node *node;
  FILE     *file;
  byte      buffer[512];

  if (!name || !*name) {
    printf ("ERROR -- no dictionary name specified\n");
    return (true);
  }
  printf ("DIC file: %s [%s]... ",name,priority ? "PRI" : "ALL");
  if (!(file = fopen(name,"r"))) { printf ("\n\nERROR: cannot open file %s\n",name); return (true); }
  while (true) {
    fgets ((char *) buffer,sizeof(buffer),file);
    if (feof(file)) break;
    lines++;
    if (!(node = new WNN_node(buffer,priority))) { printf ("\n\nERROR: Out of memory\n"); return (true); }
    if (node->type == TYPE_DELETE) { deleted++; delete node; continue; }
    if (node->type == TYPE_ERROR) return (true);
    if ((node->type == TYPE_SPECIAL) && irregular(node)) return (true);
    if (insert(node)) { dup++; delete node; continue; }
  } 
  printf ("lines=%d ent=%d total=%d del=%d dup=%d\n",lines,count-i,count,deleted,dup);
  fclose (file);
  return (false);
}

//--------------------------------
//
//  Sort the list.  This is really just a stub to call quicksort.
//
//  The sort is based on the following:
//
//      key
//      godan
//      priority
//
void WNN_index::sort () {
  printf    ("Sorting list (takes time) ... ");
  quicksort (0,count-1);
  printf    ("done\n");
  return;
}

//-------------------------------------------------------------------//
//                                                                   //
//  Man routine and satic rotuines.                                  //
//                                                                   //
//-------------------------------------------------------------------//

//--------------------------------
//
//  Gets a line of text from the configuration file.  Ending white characters are 
//  automatically stripped.
//
//      buffer -- Location to store the line in.
//      file   -- File to read from.
//
//      RETURN -- A non-zero value indicates an error.
//
int get_line (char *buffer,FILE *file) {
  if (!fgets(buffer,512,file)) return (true);
  while ((buffer[strlen(buffer)-1] == '\n') || (buffer[strlen(buffer)-1] == '\r') || (buffer[strlen(buffer)-1] == '\t') || (buffer[strlen(buffer)-1] == ' ')) buffer[strlen(buffer)-1] = 0;
  return (false);
}

//--------------------------------
//
//  A simple rotuine to get the data associated with a config file command.
//
//      buffer -- Pointer to buffer contianning the command.
//
//      RETURN -- Pointer to the data part of the command.
//
#define IS_SPACE(x)  (!((x) & 0xff00) && isspace(x))

char *get_param (char *buffer) {
  while (*buffer && !IS_SPACE(*buffer)) buffer++;
  while (*buffer &&  IS_SPACE(*buffer)) buffer++;
  return (buffer);
}

//--------------------------------
//
//  This is the main routine.
//
//  Arguments should be ovbious.
//  
int main (int argc,char **argv) {
  long      alloc = 500000;
  int       err = true;
  FILE     *file;
  WNN_index index;
  char     *ptr,buffer[512];
//
//  Basic setup.  Display message, check the parameters, and open the config file.
//
  printf ("WINFO Utility Verison 3.0 copyright (C) 2000-2002,2003 glenn rosenthal.\n\n");
  if ((argc < 2) || (argc > 3)) {
    printf ("\n\nUSSAGE: winfo <source_file> [alloc]\n");
    return (1);
  }
  if (!(file = fopen(argv[1],"r"))) {
    printf ("\n\nERROR: Cannot open the source file %s\n",argv[1]);
    return (1);
  }
  if (argc == 3) sscanf (argv[2],"%ld",&alloc);
  if (index.init(alloc)) {
    printf ("\n\nERROR: Cannot allocate index array.\n");
    return (1);
  }
//
//  Process the config file.
//
  while (!get_line(buffer,file)) {
//    printf ("[%s]\n",buffer);                                   // Debugging line.
    if (buffer[0] == 0  ) continue;                             // Blank line.
    if (buffer[0] == '#') continue;                             // Comment line.
    if (!stricmp (buffer,"end"  )) { err = false; break; }      // End of config file.
    if (!strnicmp(buffer,"wnn",3)) {                            // WNN command.  Read wnn format file.
      if (index.read_wnn(get_param(buffer))) return (1);
      continue;
    }
    if (!strnicmp(buffer,"dic",3)) {                            // DIC command.  Read EDICT format file (EUC only).
      ptr = get_param(buffer);
      if (index.read_dic(get_param(ptr),tolower(*ptr) == 'p')) return (1);
      continue;
    }
    if (!strnicmp(buffer,"wline",5)) {                          // WLINE command.  Single line in wnn format.
      if (index.read_line(get_param(buffer))) return (1);
      continue;
    }
    printf ("\n\nERROR: Unknown command in configuration [%s]\n",buffer);   // Oops!
    return (1);
  }
  fclose (file);
  if (err) {                                                    // Catches missing END command.
    printf ("\n\nERROR: Configuration file is corrupted!\n");
    return (1);
  }
//
//  Printout 
//
  index.debug  ("test1.euc");                   // Raw data.
  index.sort   ();
  index.debug  ("test2.euc");                   // Sorted
  index.filter ();
  index.debug  ("test3.euc");                   // Filtered
  if (index.merge()) return (1);
  index.debug  ("test4.euc");                   // Merged.
//  if (index.write("wnn")) return (1);           // Save older format. -- No longer realy useful, but you can enable it if you want.
  if (index.write_new("wnn")) return (1);       // Save new format.
  return (0);
}







